/*-------------------------------------------------------------------------
 *
 * replicate.h
 *	  Primary include file for replicate server .c files
 *
 * This should be the first file included by replicate modules.
 *
 *-------------------------------------------------------------------------
 */
#ifndef REPLICATE_H
#define	REPLICATE_H

#ifndef _SYS_TIME_H
#include <sys/time.h>
#endif
#include "tcop/dest.h"
#include "storage/proc.h"
#include "replicate_com.h"

#define STAND_ALONE_TAG "When_Stand_Alone"
#define NOT_REPLICATE_INFO_TAG "Not_Replicate_Info"
#define DB_NAME_TAG	"DB_Name"
#define TABLE_NAME_TAG "Table_Name"

#define CLUSTER_CONF_FILE	"cluster.conf"
#define	NOT_SESSION_AUTHORIZATION	(0)
#define SESSION_AUTHORIZATION_BEGIN	(1)
#define SESSION_AUTHORIZATION_END	(2)

#define CLUSTER_ERROR_LOG_FILE	"pgcluster.log"
#define CLUSTER_STATUS_LOG_FILE	"pgcluster.sts"
#define READ_ONLY_IF_STAND_ALONE	"read_only"
#define READ_WRITE_IF_STAND_ALONE	"read_write"
#define PERMIT_READ_ONLY		(1)
#define PERMIT_READ_WRITE		(2)
#define STATUS_REPLICATED		(3)
#define STATUS_CONTINUE			(4)
#define STATUS_CONTINUE_SELECT	(5)
#define STATUS_NOT_REPLICATE	(6)
#define STATUS_SKIP_QUERY		(7)
#define STATUS_REPLICATE		(10)
#define STATUS_RECOVERY			(11)
#define STATUS_REPLICATION_ABORT	(98)
#define STATUS_DEADLOCK_DETECT	(99)

#define TO_REPLICATION_SERVER	(0)
#define TO_FRONTEND				(1)

#define PGR_DEADLOCK_DETECTION_MSG "deadlock detected!"
#define PGR_REPLICATION_ABORT_MSG "replication aborted!"
#define SKIP_QUERY_1 "begin; select getdatabaseencoding(); commit"
#define SKIP_QUERY_2 "BEGIN; SELECT usesuper FROM pg_catalog.pg_user WHERE usename = '%s'; COMMIT"
#define SKIP_QUERY_3 "SET autocommit TO 'on'"
#define SKIP_QUERY_4 "SET search_path = public"

#define PGR_SEND_RETRY_CNT (1500)
#define PGR_SEND_WAIT_MSEC (150)
#define PGR_RECV_RETRY_CNT (1500)
#define PGR_RECV_WAIT_MSEC (150)
#define PGR_SEND_TIMEOUT	(600)
#define PGR_RECV_TIMEOUT	(30)
#define PGR_LOCK_RETRY_CNT (50)
#define PGR_LOCK_WAIT_MSEC (20)

#define PGR_1ST_RECOVERY (1)
#define PGR_2ND_RECOVERY (2)

typedef struct CopyDataType
{
	int cnt;
	char copy_data[COPYBUFSIZ];
} CopyData;

typedef struct
{
	bool is_stand_alone;
	int  permit;
} PGR_Stand_Alone_Type;

typedef struct
{
	char db_name[DBNAME_MAX_LENGTH];
	char table_name[TABLENAME_MAX_LENGTH];
} PGR_Not_Replicate_Type;

typedef struct
{
	bool check_lock_conflict;
	bool deadlock;
	int status_lock_conflict;
	int dest;
} PGR_Check_Lock_Type;

typedef struct
{
	const char * query_string;
	int query_len;
	char cmdSts;
	char cmdType;
	char useFlag;
} PGR_Retry_Query_Type;

/* cluster db data */
typedef struct {
	char * PGRSelfHostName;
	ClusterDBInfo * ClusterDBData;
	int ReplicateServerShmid;
	int ClusterDBShmid;
} PGR_ClusterDB_Info;

/* partial replication data */
typedef struct {
	int PGR_Not_Replicate_Rec_Num;
	PGR_Not_Replicate_Type * PGR_Not_Replicate;
} PGR_Partial_Info;

/* query string data */
typedef struct {
	const char * Query_String;
	const char * Query_Ptr;
	int Skip_Cnt;
	int PGR_Status;
	bool Transaction_Aborted;
	bool PGR_Not_Replication_Query;
} PGR_QueryString_Info;

/*life check process pid */
typedef struct {
	pid_t Master_Pid;
	pid_t SYN_Pid;
	pid_t ACK_Pid;
} PGR_Lifecheck_Pid_Info;

/* cluster db mode */
typedef struct {
	bool Transaction_Mode;
	bool Session_Authorization_Mode;
	bool Create_Temp_Table_Mode;
	bool PGR_Copy_Data_Need_Replicate;
	bool PGR_Reliable_Mode_Wait;
} PGR_ClusterDB_Mode_Info;

/* replication server data */
typedef struct {
	ReplicateServerInfo * ReplicateServerData;
	ReplicateServerInfo * CurrentReplicateServer;
	ReplicateServerInfo * LastReplicateServer;
	unsigned long * PGR_Current_Replication_Id;
	int TransactionSock;
	int PGR_Sock_To_Replication_Server;
} PGR_ReplicationServer_Info;

/* replicaition log */
typedef struct {
	uint32_t PGR_Replicate_ID;
	uint32_t PGR_Request_ID;
} PGR_ReplicationLog_Info;

/* Cluster DB global */
typedef struct {
	/* common data */
	PGR_Com_Info Com_Info;
	/* recovery option data */
	char * RsyncPath;
	char * RsyncOption;
	bool PGR_Recovery_Option;
	/* query buffer */
	CopyData * PGRCopyData;
	PGR_Retry_Query_Type PGR_Retry_Query;
	/* status flag */
	PGR_Stand_Alone_Type * PGR_Stand_Alone;
	bool PGR_Is_Replicated_Query;
	/* check & notice lock conflict */
	bool PGR_Need_Notice;
	bool PGR_Lock_Noticed;
	PGR_Check_Lock_Type PGR_Check_Lock;
	/* sync data */
	ReplicateNow * ReplicateCurrentTime;
} PGR_PGCluster_Info;

extern PGR_PGCluster_Info *PGCluster_Info;

/* in backend/utiles/pgcluster/replicate.c */
extern int PGR_Init_Replicate_Server_Data(char * path);
extern int PGR_Set_Replicate_Server_Socket(void);
extern char * PGR_Send_Replicate_Command(const char * query_string, int query_len, char cmdSts ,char cmdType);
extern bool PGR_Is_Replicated_Command(const char * query);
extern int Xlog_Check_Replicate(int operation);
extern int PGR_Replicate_Function_Call(void);
extern void PGR_delete_shm(void);
extern ReplicateServerInfo * PGR_get_replicate_server_info(void);
extern bool PGR_Is_System_Command(const char * query);
extern int PGR_Call_System_Command(char * command);
extern int PGR_GetTimeOfDay(struct timeval *tp, struct timezone *tpz);
extern long PGR_Random(void);
extern int PGR_Set_Current_Time(char * sec, char * usec);
extern int PGR_Send_Copy(CopyData * copy,int end);
extern CopyData * PGR_Set_Copy_Data(CopyData * copy, char * str, int len, int end);
extern char * PGR_scan_terminate(const char * str);
extern bool PGR_Is_Stand_Alone(void);
extern void PGR_Send_Message_To_Frontend(char * msg);
extern void PGR_Notice_Transaction_Query_Done(void);
extern void PGR_Notice_Transaction_Query_Aborted(void);
extern int PGR_Notice_Conflict(void);
extern int PGR_Recv_Trigger (int user_timeout);
extern bool PGR_Check_DeadLock(LOCKMETHODTABLE *lockMethodTable, LOCKMODE lockmode, LOCK *lock, PROCLOCK *holder);
extern void PGR_Set_Replication_Server_Status( ReplicateServerInfo * sp, int status);
extern int PGR_Is_Skip_Replication(const char * query);
extern bool PGR_Did_Commit_Transaction(void);
extern bool PGR_Set_Transaction_Mode(bool mode,const char * commandTag);
extern char * PGR_Remove_Comment(const char * str);
extern void PGR_Force_Replicate_Query(void);
extern void PGR_Notice_DeadLock(void);
extern void PGR_Set_Cluster_Status(int status);
extern int PGR_Get_Cluster_Status(void);
extern int PGR_lo_create(int flags);
extern int PGR_lo_open(Oid lobjId,int32 mode);
extern int PGR_lo_close(int32 fd);
extern int PGR_lo_write(int fd, char *buf, int len);
extern int PGR_lo_lseek(int32 fd, int32 offset, int32 whence);
extern int PGR_lo_unlink(Oid lobjId);
extern void PGR_init_replication(const char * query, int len);
extern void PGR_init_each_query(void);
extern int PGR_exec_replication(const char * commandTag,CommandDest dest);
extern int PGR_startup_lifecheck(void);
extern void PGR_finish_lifecheck(int sig);
extern int PGR_Wait_Reliable_Trigger(const char * commandTag);
extern void PGR_Set_Replication_Flag(Node * parsetree);
extern int PGR_Init_PGCluster_Info(void);
extern int PGR_Get_Cluster_Conf_Data(char * path);
extern void PGR_Clear_Children(int sig);
extern void PGR_Recovery_Failed(void);
extern void PGR_Elog_Abort_Copy(void);
extern void PGR_Elog_Abort_Proc(int timeout);
extern void PGR_Elog_Warn_Proc(char * msg, char * fix);
extern void PGR_ExceptionalCondition(void);
extern bool PGR_Is_This_Transaction_Committed(void);
extern void PGR_Set_Copy_Char(int ch);

/* in backend/utiles/pgcluster/recovery.c */
extern int PGR_Master_Main(void);
extern int PGR_Recovery_Main(void);
extern int PGR_recovery_error_send(void);
extern int PGR_recovery_finish_send(void);

/* in backend/access/transam/xact.c */
extern void PGR_Reload_Start_Time(void);
#endif /* REPLICATE_H */
